import pymimir.advanced.search as search

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_breadth_first_search():
    """ Test parsing a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    search_context = search.SearchContext(domain_filepath, problem_filepath)
    result = search.find_solution_brfs(search_context)

    assert result.status == search.SearchStatus.SOLVED
    assert len(result.plan) == 3



# Import all classes for better IDE support


from _pymimir.advanced.graphs import (
    # Vertices
    EmptyVertex,
    ColoredVertex,

    # Edges
    EmptyEdge,
    ColoredEdge,

    # EmptyStaticGraph
    EmptyStaticGraph,
    ImmutableEmptyStaticGraph,
    ForwardEmptyStaticGraph,
    BidirectionalEmptyStaticGraph,

    # VertexColoredStaticGraph
    VertexColoredStaticGraph,
    ImmutableVertexColoredStaticGraph,
    ForwardVertexColoredStaticGraph,
    BidirectionalVertexColoredStaticGraph,

    # EdgeColoredStaticGraph
    EdgeColoredStaticGraph,
    ImmutableEdgeColoredStaticGraph,
    ForwardEdgeColoredStaticGraph,
    BidirectionalEdgeColoredStaticGraph,
)



import pymimir.advanced.formalism as formalism
import pymimir.advanced.search as search

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


class CustomBlindHeuristic(search.IHeuristic):
    def __init__(self):
        super().__init__()

    def compute_heuristic(self, state : search.State, is_goal_state: bool) -> float:
        return 0.


class CustomGoalCountHeuristic(search.IHeuristic):
    def __init__(self, problem : formalism.Problem):
        super().__init__()  
        self.problem = problem

        self.num_goal_literals = len(self.problem.get_fluent_goal_condition()) + len(self.problem.get_derived_goal_condition())

    def compute_heuristic(self, state : search.State, is_goal_state: bool) -> float:
        num_satisfied_goal_literals = 0
        for literal in self.problem.get_fluent_goal_condition():
            if state.literal_holds(literal):
                num_satisfied_goal_literals += 1
        for literal in self.problem.get_derived_goal_condition():
            if state.literal_holds(literal):
                num_satisfied_goal_literals += 1

        return self.num_goal_literals - num_satisfied_goal_literals

class CustomEventHandler(search.IAStarEventHandler):
    def __init__(self):
        super().__init__()

    def on_expand_state(self, state : search.State):
        pass

    def on_expand_goal_state(self, state : search.State):
        pass

    def on_generate_state(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_generate_state_relaxed(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_generate_state_not_relaxed(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_close_state(self, state : search.State):
        pass

    def on_finish_f_layer(self, f_value: float):
        pass

    def on_prune_state(self, state : search.State):
        pass

    def on_start_search(self, start_state : search.State):
        pass

    def on_end_search(self, num_reached_fluent_atoms : int, num_reached_derived_atoms: int, num_bytes_for_unextended_state_portion: int, num_bytes_for_extended_state_portion: int, num_bytes_for_nodes: int, num_bytes_for_actions: int, num_bytes_for_axioms: int, num_states: int, num_nodes: int, num_actions: int, num_axioms: int):
        pass

    def on_solved(self, plan: search.Plan):
        pass

    def on_unsolvable(self):
        pass

    def on_exhausted(self):
        pass

def test_astar_search():
    """ Test parsing a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    search_context = search.SearchContext(domain_filepath, problem_filepath)

    blind_heuristic = CustomBlindHeuristic()
    goal_count_heuristic = CustomGoalCountHeuristic(search_context.get_problem())
    event_handler = CustomEventHandler()
    initial_state = search_context.get_state_repository().get_or_create_initial_state()
    default_event_handler = search.DefaultAStarEventHandler(search_context.get_problem(), False)

    result = search.find_solution_astar(search_context, blind_heuristic, initial_state, event_handler)

    assert result.status == search.SearchStatus.SOLVED
    assert len(result.plan) == 3



# Import all classes for better IDE support

from _pymimir.advanced.formalism import (
    DerivedAtom,
    DerivedAtomList,
    DerivedGroundAtom,
    DerivedGroundAtomList,
    DerivedGroundLiteral,
    DerivedGroundLiteralList,
    DerivedLiteral,
    DerivedLiteralList,
    DerivedPredicate,
    DerivedPredicateList,
    DerivedPredicateMap,
    DerivedAssignmentSet,

    FluentAtom,
    FluentAtomList,
    FluentGroundAtom,
    FluentGroundAtomList,
    FluentGroundLiteral,
    FluentGroundLiteralList,
    FluentLiteral,
    FluentLiteralList,
    FluentPredicate,
    FluentPredicateList,
    FluentPredicateMap,
    FluentAssignmentSet,

    StaticAtom,
    StaticAtomList,
    StaticGroundAtom,
    StaticGroundAtomList,
    StaticGroundLiteral,
    StaticGroundLiteralList,
    StaticLiteral,
    StaticLiteralList,
    StaticPredicate,
    StaticPredicateList,
    StaticPredicateMap,
    StaticAssignmentSet,
)

from _pymimir.advanced.formalism import (
    AuxiliaryFunction,
    AuxiliaryFunctionList,
    AuxiliaryGroundFunction,
    AuxiliaryGroundFunctionValue,
    AuxiliaryGroundFunctionValueList,
    AuxiliaryFunctionSkeleton,
    AuxiliaryFunctionSkeletonList,
    AuxiliaryNumericEffect,
    AuxiliaryFunctionExpressionFunction,
    AuxiliaryGroundFunctionExpressionFunction,
    AuxiliaryGroundNumericEffect,

    FluentFunction,
    FluentFunctionList,
    FluentGroundFunction,
    FluentGroundFunctionValue,
    FluentGroundFunctionValueList,
    FluentFunctionSkeleton,
    FluentFunctionSkeletonList,
    FluentNumericEffect,
    FluentFunctionExpressionFunction,
    FluentGroundFunctionExpressionFunction,
    FluentGroundNumericEffect,
    FluentNumericAssignmentSet,
    
    StaticFunction,
    StaticFunctionList,
    StaticGroundFunction,
    StaticGroundFunctionValue,
    StaticGroundFunctionValueList,
    StaticFunctionSkeleton,
    StaticFunctionSkeletonList,
    StaticGroundFunctionExpressionFunction,
    StaticNumericAssignmentSet,
)

from _pymimir.advanced.formalism import (
    AssignOperatorEnum,
    BinaryOperatorEnum,
    RequirementEnum,
    MultiOperatorEnum,
    OptimizationMetricEnum,
    RequirementEnum,
)

from _pymimir.advanced.formalism import (
    Action,  
    ActionList,
    Axiom,
    AxiomList,
    ConjunctiveCondition,
    ConjunctiveEffect,
    ConditionalEffect,
    ConditionalEffectList,
    Domain,
    FunctionExpression,
    FunctionExpressionList,
    FunctionExpressionBinaryOperator,
    GeneralizedProblem,
    GroundAction,
    GroundActionList,
    GroundAxiom,
    GroundAxiomList,
    GroundConjunctiveCondition,
    GroundConjunctiveEffect,
    GroundConditionalEffect,
    GroundEffectDerivedLiteral,
    GroundFunctionExpression,
    GroundFunctionExpressionList,
    GroundFunctionExpressionBinaryOperator,
    GroundFunctionExpressionMinus,
    GroundFunctionExpressionMultiOperator,
    GroundFunctionExpressionNumber,
    FunctionExpressionMinus,
    FunctionExpressionMultiOperator,
    FunctionExpressionNumber,
    Object,
    ObjectList,
    OptimizationMetric,
    Repositories,
    Parser,
    Problem,
    ProblemList,
    Requirements,
    Term,
    TermList,
    Translator,
    Variable,
    VariableList,
)



import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()

domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
problem_filepath_1 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
problem_filepath_2 = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")

search_context_options = search.SearchContextOptions()
search_context_options.mode = search.SearchMode.GROUNDED

generalized_search_context = search.GeneralizedSearchContext(domain_filepath, [problem_filepath_1, problem_filepath_2], search_context_options)

# Create KnowledgeBase
state_space_options = datasets.StateSpaceOptions()
state_space_options.symmetry_pruning = True 

generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
generalized_state_space_options.symmetry_pruning = True

knowledge_base_options = datasets.KnowledgeBaseOptions()
knowledge_base_options.state_space_options = state_space_options
knowledge_base_options.generalized_state_space_options = generalized_state_space_options

knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

# Access GeneralizedStateSpace
generalized_state_space = knowledge_base.get_generalized_state_space()

# Access GeneralizedStateSpace meta data
initial_vertices = generalized_state_space.get_initial_vertices()
unsolvable_vertices = generalized_state_space.get_unsolvable_vertices()
goal_vertices = generalized_state_space.get_goal_vertices()

# Access GeneralizedStateSpace graph
graph = generalized_state_space.get_graph()

# Iterate over vertices
for vertex in graph.get_vertices():
    # Access vertex properties (Sorry for the generic names!)
    vertex.get_index() # graph vertex index (int)
    vertex.get_property_0() # class vertex index (int)
    vertex.get_property_1() # problem index (int)

# Iterate over edges
for edge in graph.get_edges():
    # Access edge properties (Sorry for the generic names!)
    edge.get_index() # graph edge index (int)
    edge.get_source() # graph source index (int)
    edge.get_target() # graph target index (int)
    edge.get_property_0() # class edge index (int)
    edge.get_property_1() # problem index (int)

### Iteration over adjacent structures, for the example of an initial state
v_idx = initial_vertices.pop()

for vertex in graph.get_forward_adjacent_vertices(v_idx):
    pass

for vertex in graph.get_backward_adjacent_vertices(v_idx):
    pass 

for edge in graph.get_forward_adjacent_edges(v_idx):
    pass 

for edge in graph.get_backward_adjacent_edges(v_idx):
    pass 

for v2_idx in graph.get_forward_adjacent_vertex_indices(v_idx):
    pass 

for v2_idx in graph.get_backward_adjacent_vertex_indices(v_idx):
    pass 

for e2_idx in graph.get_forward_adjacent_edge_indices(v_idx):
    pass 

for e2_idx in graph.get_backward_adjacent_edge_indices(v_idx):
    pass

### To access low level information about state and actions, we can use the mappings encoded in the GeneralizedStateSpace

for vertex in graph.get_vertices():
    problem_vertex = generalized_state_space.get_problem_vertex(vertex)

    state = problem_vertex.get_property_0() # state (State)
    problem = problem_vertex.get_property_1() # problem (Problem)
    # problem_vertex.get_property_2() # certificate (optional<nauty::Certificate>)
    problem_vertex.get_property_3() # unit goal distance (int)
    problem_vertex.get_property_4() # action goal distance (double)
    problem_vertex.get_property_5() # is initial ? (bool)
    problem_vertex.get_property_6() # is goal ? (bool)
    problem_vertex.get_property_7() # is unsolvable ? (bool)
    problem_vertex.get_property_8() # is alive (bool)

    problem = generalized_state_space.get_problem(vertex)

    print(state.to_string(problem))

for edge in graph.get_edges():
    problem_edge = generalized_state_space.get_problem_edge(edge)

    action = problem_edge.get_property_0() # action (GroundAction)
    problem = problem_edge.get_property_1() # problem (Problem)
    action_cost = problem_edge.get_property_2() # cost (double)

    problem = generalized_state_space.get_problem(edge)

    print(action.to_string(problem))




# Enums
from _pymimir.advanced.languages.description_logics import (
    GrammarSpecificationEnum,
)

# DL constructors
from _pymimir.advanced.languages.description_logics import (
    ConceptConstructor,
    ConceptConstructorList,
    RoleConstructor,
    RoleConstructorList,
    BooleanConstructor,
    BooleanConstructorList,
    NumericalConstructor,
    NumericalConstructorList,

    ConceptBotConstructor,
    ConceptTopConstructor,

    Repositories,

    ConstructorVisitor,
)

# CNF grammar

from _pymimir.advanced.languages.description_logics import (
    CNFGrammar,
)


import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_knowledge_base_without_symmetry_reduction():
    """ Test the construction of generalized state space
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath_1 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
    problem_filepath_2 = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")

    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = False 

    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    generalized_state_space_options.symmetry_pruning = False

    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options

    generalized_search_context = search.GeneralizedSearchContext(domain_filepath, [problem_filepath_1, problem_filepath_2])

    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    generalized_state_space = knowledge_base.get_generalized_state_space()

    class_graph = generalized_state_space.get_graph()
    assert class_graph.get_num_vertices() == 36
    assert class_graph.get_num_edges() == 128
    assert len(generalized_state_space.get_goal_vertices()) == 4
    assert len(generalized_state_space.get_unsolvable_vertices()) == 0

def test_knowledge_base_with_symmetry_reduction():
    """ Test the construction of generalized state space
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath_1 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
    problem_filepath_2 = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")

    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = True 

    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    generalized_state_space_options.symmetry_pruning = True

    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options

    generalized_search_context = search.GeneralizedSearchContext(domain_filepath, [problem_filepath_1, problem_filepath_2])

    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    generalized_state_space = knowledge_base.get_generalized_state_space()

    class_graph = generalized_state_space.get_graph()
    assert class_graph.get_num_vertices() == 18
    assert class_graph.get_num_edges() == 52
    assert len(generalized_state_space.get_goal_vertices()) == 4
    assert len(generalized_state_space.get_unsolvable_vertices()) == 0




# Import all classes for better IDE support

from _pymimir.advanced.datasets import (
    # Vertices
    ProblemVertex,
    ProblemEdge,

    # Edges
    ClassVertex,
    ClassEdge,

    # ProblemStaticGraph
    ProblemStaticGraph,
    ImmutableProblemStaticGraph,
    ForwardProblemStaticGraph,
    BidirectionalProblemStaticGraph,

    # ClassStaticGraph
    ClassStaticGraph,
    ImmutableClassStaticGraph,
    ForwardClassStaticGraph,
    BidirectionalClassStaticGraph,

    # StateSpace
    StateSpace,
    StateSpaceOptions,

    # GeneralizedStateSpace
    GeneralizedStateSpace,
    GeneralizedStateSpaceOptions,

    # TupleGraph
    TupleGraphOptions,

    # KnowledgeBase
    KnowledgeBase,
    KnowledgeBaseOptions,
)

__all__ = [
    "ProblemVertex",
    "ProblemEdge",
    "ClassVertex",
    "ClassEdge",

    "ProblemStaticGraph",
    "ImmutableProblemStaticGraph",
    "ForwardProblemStaticGraph",
    "BidirectionalProblemStaticGraph",

    "ClassStaticGraph",
    "ImmutableClassStaticGraph",
    "ForwardClassStaticGraph",
    "BidirectionalClassStaticGraph",

    "GeneralizedStateSpace",
    "GeneralizedStateSpaceOptions",
    "GeneralizedStateSpaceProblemOptions",

    "KnowledgeBase",
    "KnowledgeBaseOptions",
    "TupleGraphCollectionOptions",
]


import pymimir.advanced.formalism as formalism
import pymimir.advanced.search as search

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_applicable_action_generator_ownership():
    """ Test `ApplicableActionGenerator`.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    problem = formalism.Problem.create(domain_filepath, problem_filepath)

    applicable_action_generator = search.LiftedApplicableActionGenerator(problem)
    axiom_evaluator = search.LiftedAxiomEvaluator(problem)
    state_repository = search.StateRepository(axiom_evaluator)
    initial_state = state_repository.get_or_create_initial_state()
    actions = applicable_action_generator.generate_applicable_actions(initial_state)

    assert len(actions) == 6
    assert actions[0].to_string(problem)
